Package: seqmagick
Title: Sequence Manipulation Utilities
Version: 0.1.9
Authors@R: c(
    person("Guangchuang", "Yu",     email = "guangchuangyu@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-6485-8781"))
    )
Description: Tools for reading and writing biological sequences in multiple
    formats, including 'FASTA', 'PHYLIP', 'CLUSTAL', 'STOCKHOLM', 'MEGA' and
    'GenBank'. Supports interleaved and sequential layouts where applicable,
    and reads or writes gzip, bzip2 and xz compressed files transparently.
    Converts between formats, renames sequences via a two-column mapping,
    summarizes sequence counts, lengths, GC content and ambiguous characters,
    and manipulates sequence sets (e.g., filtering by patterns and computing
    consensus sequences from alignments). Also includes functions to download
    nucleotide records from NCBI by accession.
Depends: R (>= 4.1.0)
Imports: Biostrings, stats, utils, yulab.utils (>= 0.2.5)
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), GenomicAlignments,
        GenomicRanges, IRanges, muscle, Rsamtools, prettydoc
VignetteBuilder: knitr
ByteCompile: true
URL: https://github.com/YuLab-SMU/seqmagick
BugReports: https://github.com/YuLab-SMU/seqmagick/issues
License: Artistic-2.0
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-08-30 06:12:08 UTC; wang
Author: Guangchuang Yu [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6485-8781>)
Maintainer: Guangchuang Yu <guangchuangyu@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-30 06:30:02 UTC

# Generated by roxygen2: do not edit by hand

export(bam2DNAStringSet)
export(bam2DNAStringSet2)
export(bs_aln)
export(bs_consensus)
export(bs_filter)
export(bs_hamming)
export(bs_rename)
export(clw_read)
export(consensus)
export(download_genbank)
export(fa_combine)
export(fa_filter)
export(fa_read)
export(fa_rename)
export(fa_to_interleaved)
export(fa_to_sequential)
export(fa_write)
export(fas2phy)
export(gb_read)
export(get_id)
export(mega_read)
export(ncbi_fa_read)
export(phy2fas)
export(phy_read)
export(phy_write)
export(renameTXT)
export(replaceInside)
export(seqlen)
export(sth_read)
importFrom(Biostrings,AAStringSet)
importFrom(Biostrings,BStringSet)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,RNAStringSet)
importFrom(Biostrings,consensusMatrix)
importFrom(Biostrings,matchPattern)
importFrom(Biostrings,readAAStringSet)
importFrom(Biostrings,readBStringSet)
importFrom(Biostrings,readDNAStringSet)
importFrom(Biostrings,readRNAStringSet)
importFrom(Biostrings,toString)
importFrom(Biostrings,width)
importFrom(Biostrings,writeXStringSet)
importFrom(magrittr,"%<>%")
importFrom(magrittr,"%>%")
importFrom(stats,setNames)
importFrom(utils,download.file)
importFrom(utils,read.delim)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
importFrom(yulab.utils,use_perl)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# seqmagick: Sequence Manipulation Utilities

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/seqmagick)](https://cran.r-project.org/package=seqmagick)
[![CRAN_Download_Badge](https://cranlogs.r-pkg.org/badges/grand-total/seqmagick?color=green)](https://cran.r-project.org/package=seqmagick)
<!-- badges: end -->

Tools for reading and writing biological sequences in multiple formats,
including ‘FASTA’, ‘PHYLIP’, ‘CLUSTAL’, ‘STOCKHOLM’, ‘MEGA’ and
‘GenBank’. Supports interleaved and sequential layouts where applicable,
and reads or writes gzip, bzip2 and xz compressed files transparently.
Converts between formats, renames sequences via a two-column mapping,
summarizes sequence counts, lengths, GC content and ambiguous
characters, and manipulates sequence sets (e.g., filtering by patterns
and computing consensus sequences from alignments). Also includes
functions to download nucleotide records from NCBI by accession.

## :writing_hand: Authors

YuLab@SMU <https://yulab-smu.top>

## :page_facing_up: Supported file formats

| Format                  | Read               | Write              |
|:------------------------|:-------------------|:-------------------|
| FASTA (.fas/.fa/.fasta) | :heavy_check_mark: | :heavy_check_mark: |
| PHYLIP (.phy)           | :heavy_check_mark: | :heavy_check_mark: |
| CLUSTAL (.clw)          | :heavy_check_mark: | \-                 |
| STOCKHOLM (.sth)        | :heavy_check_mark: | \-                 |
| MEGA (.meg)             | :heavy_check_mark: | \-                 |
| GenBank (.gb)           | :heavy_check_mark: | \-                 |
| NCBI (online)           | :heavy_check_mark: | \-                 |

Both sequential and interleaved layouts are supported for FASTA and
PHYLIP. Compressed files (`gzip`, `bzip2`, `xz`) are handled
transparently.

## :hammer: Installation

Install from CRAN:

``` r
install.packages("seqmagick")
```

Or grab the development version from GitHub:

``` r
if (!requireNamespace("remotes", quietly=TRUE))
    install.packages("remotes")
remotes::install_github("YuLab-SMU/seqmagick")
```

## :bulb: Quick Examples

``` r
library(seqmagick)

fas <- system.file("extdata/HA.fas", package = "seqmagick")

## reading/writing FASTA files
x <- fa_read(fas)
fa_write(x[1:5], "sub.fa.gz")     # gzip/bzip2/xz supported

## a quick summary (counts, lengths, GC content)
fa_summary(fas)

## sequence manipulation
bs_filter(x, 'ATGAAAGTAAAA', by='sequence')   # keep matching sequences

## renaming sequences via two-column mapping
map <- data.frame(old = names(x)[1:2],
                  new = c("HA_HK", "HA_BR"))
y <- bs_rename(x, map)

## format conversion (requires aligned sequences)
fas2phy("aln.fas", "aln.phy")
phy2fas("aln.phy", "aln.fas")

## download nucleotide records from NCBI
seqs <- ncbi_fa_read(acc = c("AB115403", "CY084969"))
```

For more details, please see the [online
vignette](https://CRAN.R-project.org/package=seqmagick).

If you use `seqmagick`, please cite:

> Yu G (2026). *seqmagick: Sequence Manipulation Utilities*. R package
> version 0.1.9, <https://github.com/YuLab-SMU/seqmagick>.

<br>

## ----style, echo=FALSE, results="asis", message=FALSE-------------------------
knitr::opts_chunk$set(tidy = FALSE,
		   message = FALSE)

## ----echo=FALSE, results="hide", message=FALSE--------------------------------
library(magrittr)
library(Biostrings)
library("seqmagick")

## ----eval=FALSE---------------------------------------------------------------
# tmpgb <- tempfile(fileext = '.gb')
# tmpfa <- tempfile(fileext = '.fa')
# download_genbank(acc='AB115403', format='genbank', outfile=tmpgb)
# download_genbank(acc='AB115403', format='fasta', outfile=tmpfa)
# ## readLines(tmpgb)[1:10]
# ## readLines(tmpfa)

## ----eval=F-------------------------------------------------------------------
# fa_file <- system.file("extdata/HA.fas", package="seqmagick")
# ## use the small subset to save compilation time of the vignette
# fa2 <- tempfile(fileext = '.fa')
# fa_read(fa_file) %>% bs_filter('ATGAAAGTAAAA', by='sequence') %>% fa_write(fa2, type='interleaved')
# 
# 
# alnfas <- tempfile(fileext = ".fas")
# fa_read(fa2) %>% bs_aln(quiet=TRUE) %>% fa_write(alnfas)
# 
# ## phylip format is only for aligned sequences
# tmpphy <- tempfile(fileext = ".phy")
# fas2phy(alnfas, tmpphy, type = 'sequential')

## ----eval=F-------------------------------------------------------------------
# phy2fas(tmpphy, alnfas, type = 'interleaved')

## ----eval=F-------------------------------------------------------------------
# tmpfas <- tempfile(fileext='.fa')
# fa_read(fa2) %>% fa_write(tmpfas, type="sequential")

## ----eval=F-------------------------------------------------------------------
# tmpphy2 <- tempfile(fileext = '.phy')
# phy_read(tmpphy) %>% phy_write(tmpphy2, type="interleaved")

## ----eval=F-------------------------------------------------------------------
# bs <- fa_read(fa_file)
# bs_filter(bs, 'ATGAAAGTAAAA', by='sequence')
# 
# aln <- bs_filter(bs, 'ATGAAAGTAAAA', by='sequence') %>% bs_aln(quiet=TRUE)
# 
# bs_consensus(aln)

## ----echo=FALSE---------------------------------------------------------------
sessionInfo()

##' write fasta file
##'
##' 
##' @title fa_write
##' @param x XStringSet object
##' @param outfile output file
##' @param type one of interleaved and sequential
##' @return None
##' @export
##' @author Guangchuang Yu
##' @references \url{https://phylipweb.github.io/phylip/}
##' @examples
##' phy_file <- system.file("extdata/HA.phy", package="seqmagick")
##' x <- phy_read(phy_file)
##' fa_file <- tempfile(fileext = '.fas')
##' fa_write(x, fa_file)
fa_write <- function(x, outfile, type="interleaved") {
    type <- match.arg(type, c("interleaved", "sequential"))
    if (type == "interleaved") {
        fa_write_interleaved(x, outfile)
    } else {
        fa_write_sequential(x, outfile)
    }
}

##' write phylip file
##'
##' 
##' @title phy_write
##' @param x XStringSet object
##' @param outfile output file
##' @param type one of interleaved and sequential 
##' @return None
##' @export
##' @author Guangchuang Yu
##' @examples
##' \dontrun{
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' aln <- bs_aln(x[1:5])
##' phy_file <- tempfile(fileext = '.phy')
##' phy_write(aln, phy_file)
##' }
phy_write <- function(x, outfile, type="sequential") {
    type <- match.arg(type, c("interleaved", "sequential"))

    if (type == "interleaved") {
        phy_write_interleaved(x, outfile)
    } else {
        phy_write_sequential(x, outfile)
    }
}

check_alnseq <- function(x) {
    if (length(unique(width(x))) != 1) {
        stop("input fasfile should contains aligned sequences with equal length...")
    }
}

##' @importFrom Biostrings width
##' @importFrom Biostrings toString
phy_write_sequential <- function(x, outfile) {
    check_alnseq(x)

    w <- width(x[1])
    out <- file(outfile, "w")
    header <- paste(length(x), "\t", w)
    writeLines(header, out)
    nn <- max(nchar(names(x)))
    for (i in 1:length(x)) {
        n <- names(x[i])
        sep.blank <- paste(rep(" ", nn-nchar(n)+4), sep="", collapse="")
        line <- paste(n, sep.blank, toString(x[i]), sep="")
        writeLines(line, out)
    }
    close(out)
}

phy_write_interleaved <- function(x, outfile) {
    check_alnseq(x)
    
    w <- width(x[1])
    out <- file(outfile, "w")
    header <- paste(length(x), "\t", w)
    writeLines(header, out)
    nn <- max(nchar(names(x)))
    for (j in 1:ceiling(w/50)) {
        for (i in seq_along(x)) {
            if (j == 1) {
                n <- names(x[i])
                sep.blank <- paste(rep(" ", nn-nchar(n)+4), sep="", collapse="")
                preceding <- paste(n, sep.blank)
            } else {
                preceding <- paste(rep(" ", nn + 5), sep="", collapse="")
            }
            
            to <- seq(j*5-4, j*5) * 10
            from <- to - 9
            if (to[5] > w) {
                idx <- which(to >= w)[1]
                from <- from[1:idx]
                to <- to[1:idx]
                to[idx] <- w
            }
            s <- paste(substring(toString(x[i]), from, to), collapse=" ")
            line <- paste0(preceding, s)
            writeLines(line, out)
        }
        if (j != ceiling(w/50)) {
            linebreak=""
            writeLines(linebreak, out)
        }
    }
    close(out)
}

fa_write_interleaved <- function(x, outfile) {
    out <- .seq_connection(outfile, "w")
    on.exit(close(out))
    seqs <- as.character(x)
    w <- 60L
    for (i in seq_along(seqs)) {
        writeLines(paste0(">", names(x)[i]), out)
        n <- nchar(seqs[i])
        if (n <= w) {
            writeLines(seqs[i], out)
        } else {
            from <- seq(1L, n, by=w)
            to <- pmin(from + w - 1L, n)
            writeLines(substring(seqs[i], from, to), out)
        }
    }
}

fa_write_sequential <- function(x, outfile) {
    out <- .seq_connection(outfile, "w")
    desc <- names(x)

    for (i in seq_along(desc)) {
        writeLines(paste0('>', desc[i]), out)
        writeLines(toString(x[i]), out)
    }
    
    close(out)
}

fas <- system.file("extdata/HA.fas", package="seqmagick")
phy <- system.file("extdata/HA.phy", package="seqmagick")

test_that("fa_read returns expected XStringSet", {
    x <- fa_read(fas)
    expect_s4_class(x, "XStringSet")
    expect_equal(length(x), 100)
})

test_that("fa_read type='auto' vs explicit type agree on DNA file", {
    a <- fa_read(fas)
    b <- fa_read(fas, type="DNA")
    expect_identical(as.character(a), as.character(b))
})

## compression formats x writing styles
test_that("fa_write/fa_read round-trip across compressed formats", {
    x <- fa_read(fas)[1:10]
    for (type in c("interleaved", "sequential")) {
        for (ext in c(".fas", ".fas.gz", ".fasta.bz2", ".fasta.xz")) {
            f <- tempfile(fileext=ext)
            fa_write(x, f, type=type)
            y <- fa_read(f)
            info <- paste(type, ext)
            expect_identical(names(y), names(x), info=info)
            expect_identical(as.character(y), as.character(x), info=info)
            unlink(f)
        }
    }
})

test_that("phylip round-trip (sequential)", {
    p <- phy_read(phy)
    f <- tempfile(fileext=".phy")
    phy_write(p, f, type="sequential")
    q <- phy_read(f)
    expect_identical(names(q), names(p))
    expect_identical(unname(as.character(q)), unname(as.character(p)))
    unlink(f)
})

test_that("phylip round-trip (interleaved)", {
    p <- phy_read(phy)
    f <- tempfile(fileext=".phy")
    phy_write(p, f, type="interleaved")
    q <- phy_read(f)
    expect_identical(names(q), names(p))
    expect_identical(gsub("-", "", unname(as.character(q))),
                     gsub("-", "", unname(as.character(p))))
    unlink(f)
})

test_that("phylip -> fasta conversion keeps sequences", {
    p <- phy_read(phy)[1:5]
    f <- tempfile(fileext=".fas")
    fa_write(p, f)
    x <- fa_read(f)
    expect_identical(names(x), names(p))
    expect_identical(unname(as.character(x)), unname(as.character(p)))
    unlink(f)
})

#' @rdname msa-read
#' @export
#' @importFrom yulab.utils use_perl
mega_read <- function(file) {
    con <- .seq_connection(file, "r")
    on.exit(close(con))
    x <- readLines(con)
    start <- grep("^#", x, perl=use_perl())[2] - 1
    header <- x[1:start]
    format <- mega_format(header)

    if (is.null(format)) {
        res <- mega_record_no_format(x[-(1:start)])
        return(res)
    }

    i <- grep("^!Gene=", x, perl=use_perl())

    if (length(i) == 0) {
        records <- x[-(1:start)]

        res <- mega_record(records, format)
        return(res)
    } 
    genename <- sub("!Gene=", "", x[i], perl=use_perl())
    genename <- sub(";", "", genename, perl=use_perl())
    genename <- sub("\\s+.*", "", genename, perl=use_perl())

    begin <- i + 1
    end <- c(i[-1] -1, length(x))

    res <- lapply(seq_along(genename), function(j) {
        records <- x[begin[j]:end[j]]
        mega_record_no_format(records, format["DataType"])
    })
    names(res) <- genename
    return(res)
}

mega_record <- function(records, format) {
    records <-  sub("^#", ">", records, perl=use_perl())
    records <- gsub("\\s+", "", records, perl=use_perl())

    ic <- format['identical']
    type <- format["DataType"]
    if (is.na(type)) {
        type <- guess_sequence_type(records[2])
    }

    f <- tempfile(fileext = ".fa")
    cat(file = f, records, sep="\n")

    if (is.na(ic)) {
        res <- fa_read(f, type)
        return(res)
    }

    bs <- readBStringSet(f)

    bsm <- as.matrix(bs)
    for (i in 2:nrow(bsm)) {
        j = which(bsm[i, ] == ic) 
        bsm[i, j] = bsm[1, j]
    }
    seq_str <- apply(bsm, 1, paste0, collapse='') 

    build_seq_object(seq_str, type)
}

mega_record_no_format <- function(records, type = NA) {
    x <- parse_name_seq(records)
    seqname <- sub("^#", "", x$seqname, perl=use_perl())
    seqs <- collapse_seqs(seqname, x$seqs)
    
    build_seq_object(seqs, type)
}

collapse_seqs <- function(seqname, seqs) {
    if (any(duplicated(seqname))) {
        seqs <- vapply(split(seqs, seqname), paste0, collapse = "", FUN.VALUE=character(1))
    } else {
        names(seqs) <- seqname
    }
    return(seqs)
}

parse_name_seq <- function(records) {
    records <- records[records != ""]
    seqname <- sub("^(\\S+)\\s+.*", "\\1", records, perl=use_perl())
    seqs <- sub("^\\S+", "", records, perl=use_perl())
    seqs <- gsub("\\s+", "", seqs, perl=use_perl())
    return(list(seqname = seqname, seqs = seqs))
}


build_seq_object <- function(seq_str, type) {
    if (is.na(type) || type == "auto") {
        type <- guess_sequence_type(seq_str[1])
    }

    type <- toupper(type)
    
    switch(type,
           DNA = DNAStringSet(seq_str),
           RNA = RNAStringSet(seq_str),
           AA  = AAStringSet(seq_str),
           PROTEIN  = AAStringSet(seq_str),
           UNKNOWN = BStringSet(seq_str)
           )    
}

mega_format <- function(header) {
    i <- grep("!Format\\s+DataType=", header, perl=use_perl())
    if (length(i) == 0) {
        # not format definition
        return(NULL)
    } 

    f <- gsub("!Format\\s*|;", "", header[i], perl=use_perl())
    ff <- strsplit(strsplit(f, " ")[[1]], "=") 
    fm <- do.call(rbind, ff)
    res <- setNames(fm[,2], fm[,1])
    return(res)
}

##' read sequence alignment file
##'
##'
##' @rdname msa-read
##' @param file multiple sequence file, gzip/bzip2/xz compressed files are supported (.gz/.bz2/.xz)
##' @param type one of 'DNA', 'RNA', 'AA', 'Protein', 'unknown' or 'auto'
##' @return BStringSet object
##' @importFrom Biostrings readBStringSet
##' @importFrom Biostrings readDNAStringSet
##' @importFrom Biostrings readRNAStringSet
##' @importFrom Biostrings readAAStringSet
##' @export
##' @author Guangchuang Yu
##' @examples
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' fa_read(fa_file)
##' 
##' mega_file <- system.file("extdata/mega/Crab_rRNA.meg", package="seqmagick")
##' mega_read(mega_file)
fa_read <- function(file, type = "auto") {
    file <- .plain_text_file(file)
    type <- match.arg(type, c("DNA", "RNA", "AA", "Protein", "unknown", "auto" ))
    if (type == "auto") {
        type <- guess_sequence_type(file)
    }
    type <- toupper(type)
    switch(type,
           DNA = readDNAStringSet(file),
           RNA = readRNAStringSet(file),
           AA  = readAAStringSet(file),
           PROTEIN  = readAAStringSet(file),
           UNKNOWN = readBStringSet(file)
           )
}


##' read aligned sequences in phylip format
##'
##'
##' @title phy_read
##' @param file phylip file
##' @importFrom Biostrings BStringSet
##' @return BStringSet object
##' @export
##' @author Guangchuang Yu
##' @examples
##' phy_file <- system.file("extdata/HA.phy", package="seqmagick")
##' phy_read(phy_file)
phy_read <- function(file) {
    info <- getPhyInfo(file)
    phy <- readLines(file)

    type <- "interleaved"
    if (length(phy) == (info$num + 1)) {
        type <- "sequential"
    }

    if (type == "sequential") {
        phy <- phy[-1]

        ii <- nchar(phy) - info$width
        nm <- gsub("\\s+$", "", substring(phy, 1,ii-1))
        seq_str <- substring(phy, ii+1, nchar(phy))

        ## may use \t and lines are not in equal length
        ## nm <- gsub("\\s+[^\\s]+$", "", phy)
        ## seq_str <- gsub("^[^\\s]+\\s+", "", phy)

    } else {
        phy <- phy[nchar(phy) != 0]
        phy <- phy[-1]
        ii <- regexpr("\\S", phy[info$num+2])
        nn <- nchar(phy[info$num+2])
        nm <- gsub("\\s+$", "", substring(phy[1:info$num], 1,ii-1))
        ss <- substring(phy, ii, nn)

        seq_str <- sapply(1:info$num, function(i) {
            j <- seq(i, length(ss), by=info$num)
            gsub(" ", "", paste0(ss[j], sep="", collapse=""))
        })
    }

    names(seq_str) <- nm
    ## BStringSet(seq_str)
    switch(guess_sequence_type(seq_str[1]),
           DNA = DNAStringSet(seq_str),
           RNA = RNAStringSet(seq_str),
           AA = AAStringSet(seq_str))
}

guess_sequence_type <- function(string) {
    if (file.exists(string)){
        con <- .seq_connection(string, "r")
        on.exit(close(con))
        seqstr <- readLines(con, n=20)
        if (length(seqstr)==2 || grepl("^>", seqstr[[3]])){
            # >seq1
            # AGCGTACGTGACGTAGCGTAGC
            # >seq2
            a <- seqstr[2]
        }else{
            # >seq1
            # ---------
            # AGCG----C
            # ---------
            # >seq2
            seqind <- grep("^>", seqstr)
            if (length(seqind) < 2){
                a <- paste0(seqstr[-1], collapse="")
            }else{
                inds <- seqind[1] + 1
                inde <- seqind[2] - 1
                a <- paste0(seqstr[inds:inde], collapse="")
            }
        }
    }else{
        a <- string
    }
    a <- strsplit(toupper(a), split="")[[1]]
    freq1 <- mean(a %in% c('A', 'C', 'G', 'T', 'X', 'N', '-') )
    freq2 <- mean(a %in% c('A', 'C', 'G', 'T', 'N'))
    if (freq1 > 0.9 && freq2 > 0) {
        return('DNA')
    }

    freq3 <- mean(a %in% c('A', 'C', 'G', 'U', 'X', 'N', '-'))
    freq4 <- mean(a %in% c('T'))
    freq5 <- mean(a %in% c('U'))
    if (freq3 > 0.9 && freq4==0 && freq5 > 0) {
        return('RNA')
    }

    return('AA')
}



##' extract accession number and sequence from genbank file
##'
##' 
##' @title gb_read
##' @param file input genbank file
##' @return sequence object
##' @export
##' @author Guangchuang Yu
gb_read <- function(file) {
    res <- vapply(file, gb_read_item, FUN.VALUE = character(1))
    
    f <- tempfile(fileext = ".fasta")
    cat(res, file = f, sep="")
    fa_read(f)
}

gb_read_item <- function(file) {
    x <- readLines(file)
    acc <- sub("\\w+\\s+(\\w+)$", "\\1", x[grep("^ACCESSION", x)])
    src <- sub("SOURCE\\s+",  "", x[grep("^SOURCE",x)])
    header <- paste0(src, "(", acc, ")")
    i <- grep("ORIGIN", x)
    ss <- x[(i+1):length(x)]
    ss <- ss[1:(grep("//", ss) -1)]
    ss <- gsub("\\s+\\d+", "", ss)
    ss <- gsub("\\s+", "", ss)
    ss <- paste0(ss, collapse = "")

    # f <- tempfile(fileext = ".fasta")
    # cat(">",  file = f, append = TRUE)
    # cat(header,  file = f, append = TRUE, sep = "\n")
    # cat(ss,  file = f, append = TRUE, sep = "\n")
    # fa_read(f)

    sprintf(">%s\n%s\n", header, ss)
}




#' @rdname msa-read
#' @export
clw_read <- function(file, type = "auto") {
    x <- readLines(file)
    program <- toupper(sub("(\\w+)\\s.*", "\\1", x[1], perl=use_perl()))
    if (!program %in% c("CLUSTAL", "MUSCLE")) {
        stop("wrong file format")
    }
    i <- grep("^\\s+", x, perl=use_perl())
    x <- x[-c(1, i)]
    
    y <- parse_name_seq(x)
    y <- collapse_seqs(y$seqname, y$seqs)
    build_seq_object(y, type)
}

#' @rdname msa-read
#' @export
#' @importFrom stats setNames
sth_read <- function(file, type = "auto") {
    # STOCKHOLM file
    con <- .seq_connection(file, "r")
    on.exit(close(con))
    x <- readLines(con)
    if (!grepl("STOCKHOLM", x[1], perl=use_perl())) {
        stop("wrong file format")
    }

    seqname <- x[grep("^#=GS", x, perl=use_perl())]
    seqname <- sub("#=GS\\s(\\S+)\\s.*", "\\1", seqname, perl=use_perl())

    key <- sub("^(\\S+)\\s.*", "\\1", x, perl=use_perl())
    value <- sub("^\\S+\\s+", "", x, perl=use_perl())
    i <- which(key %in% seqname)
    seqs <- setNames(value[i], key[i])
    build_seq_object(seqs, type)
}

write_mini_fasta <- function(seqs, per_line=FALSE) {
    f <- tempfile(fileext=".fas")
    con <- file(f, "w")
    for (i in seq_along(seqs)) {
        writeLines(paste0(">seq", i), con)
        if (per_line) {
            half <- nchar(seqs[i]) %/% 2
            writeLines(substring(seqs[i], 1, half), con)
            writeLines(substring(seqs[i], half+1), con)
        } else {
            writeLines(seqs[i], con)
        }
    }
    close(con)
    f
}

test_that("guess_sequence_type infers DNA/RNA/AA from raw string", {
    expect_equal(seqmagick:::guess_sequence_type("ACGTAGCTAN"), "DNA")
    expect_equal(seqmagick:::guess_sequence_type("ACGUAGCUAN"), "RNA")
    expect_equal(seqmagick:::guess_sequence_type("MAEGKLVIWF"), "AA")
})

test_that("fa_read type='auto' detects sequence types end-to-end", {
    dna <- write_mini_fasta(c("ACGTNNAGCTTA-", "TTTAGCCAN"))
    rna <- write_mini_fasta(c("ACGUUNAGCUUA-", "UUUAGCCAN"))
    aa  <- write_mini_fasta(c("MAEGKLVIWFKLM", "MWKFTESTQNLPP"))

    expect_s4_class(fa_read(dna), "DNAStringSet")
    expect_s4_class(fa_read(rna), "RNAStringSet")
    expect_s4_class(fa_read(aa),  "AAStringSet")

    unlink(c(dna, rna, aa))
})

test_that("guess_sequence_type handles multi-line fasta records", {
    ## interleaved-style records where one sequence spans several lines
    f <- write_mini_fasta(c("ACGTACGTACGT", "TTTACGTAAAAC"), per_line=TRUE)
    expect_s4_class(fa_read(f, type="auto"), "DNAStringSet")
    unlink(f)
})

build_fas <- function(seqs) {
    f <- tempfile(fileext=".fas")
    con <- file(f, "w")
    for (i in seq_along(seqs)) {
        writeLines(paste0(">s", i), con)
        writeLines(seqs[i], con)
    }
    close(con)
    f
}

test_that("fa_summary computes counts, lengths, GC and ambiguity", {
    f <- build_fas(c("ACGT", "ACGN", "GT--"))
    on.exit(unlink(f))

    res <- fa_summary(f)

    expect_s3_class(res, "data.frame")
    expect_identical(dim(res), c(1L, 6L))
    expect_identical(names(res),
                     c("n_seq", "min_len", "max_len", "mean_len",
                       "GC_pct", "ambiguous_pct"))

    expect_equal(res$n_seq, 3)
    expect_equal(res$min_len, 4)
    expect_equal(res$max_len, 4)
    expect_equal(res$mean_len, 4)

    ## 12 chars total: 9 clean (A/C/G/T/U), of which 5 are G or C
    expect_equal(res$GC_pct, round(100 * 5/9, 2))
    ## 3 ambiguous: N, -, -
    expect_equal(res$ambiguous_pct, 25)
})

test_that("fa_summary accepts XStringSet input directly", {
    f <- build_fas("ACGTRYKMS")
    x <- fa_read(f, type="unknown")

    res_file <- fa_summary(f)
    res_obj  <- fa_summary(x)
    expect_identical(res_file, res_obj)

    unlink(f)
})

##' consensus of aligned sequences
##'
##' 
##' @title consensus
##' @rdname consensus
##' @param x XStringSet or XMultipleAlignment object
##' @param type currently, only DNA supported
##' @return consensus sequence string
##' @export
##' @author Guangchuang Yu
##' @examples
##' \dontrun{
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' ## align first 5 sequences, use `bs_aln(x)` to align all sequences
##' aln <- bs_aln(x[1:5])
##' ## or bs_consensus(aln)
##' consensus(aln)
##' }
consensus <- function(x, type="DNA") {
    bs_consensus(BStringSet(x))
}

##' consensus of aligned sequences
##'
##' 
##' @title bs_consensus
##' @rdname consensus
##' @param x BStringSet object
##' @param type currently, only DNA supported
##' @param r if any NT > r, it will be selected as representative base
##' @return consensus sequence string
##' @importFrom Biostrings width
##' @importFrom Biostrings consensusMatrix
##' @export
##' @author Guangchuang Yu
bs_consensus <- function(x, type="DNA", r=1) {
    if (length(unique(width(x))) != 1) {
        stop("input sequences were not aligned...")
    }
    
    if (type != "DNA")
        stop("currently only DNA is supported...")
    
    y <- consensusMatrix(x)
    cns <- rep(NA, ncol(y))

    if (type == "DNA") {
        alphabet <- c("A", "C", "G", "T")
        ambiguity_code <- DNA_ambiguity_code
    }
    
    idx <- match(alphabet, rownames(y))
    nn <- apply(y[idx,], 2, function(n) sum(n != 0))
    i <- which(nn == 1)
    j <- apply(y[idx, i], 2, function(n) which(n!=0))
    cns[i] <- alphabet[j]
    
    cns[nn == 0] <- '-'
    
    if (sum(nn > 1)> 0) {
        ambiguity_alphabet <- apply(y[1:4, nn > 1, drop=FALSE], 2, function(n) {
            rr <- n/sum(n)
            if (any(rr > r)) {
                ii <- which(rr > r)
            } else {
                ii <- which(rr > 0)
            }
            ambiguity_code(alphabet[ii])
        })
        cns[nn > 1] <- ambiguity_alphabet
    }
    paste0(cns, collapse='')
}

DNA_ambiguity_code <- function(NT) {
    ## M          A or C
    ## R          A or G
    ## W          A or T
    ## S          C or G
    ## Y          C or T
    ## K          G or T
    ## V        A or C or G
    ## H        A or C or T
    ## D        A or G or T
    ## B        C or G or T
    ## X      G or A or T or C
    ## N      G or A or T or C

    NT <- NT |> unique() |>
        toupper() |>
        base::sort()
    
    if (length(NT) < 2) {
        return(NT)
    }

    if (length(NT) > 4) {
        stop("length of 'NT' should be ranged between 2 and 4...")
    }

    if (length(NT) == 2) {
        if (all(NT == c("A", "C")))
            return("M")
        if (all(NT == c("A", "G")))
            return("R")
        if (all(NT == c("A", "T")))
            return("W")
        if (all(NT == c("C", "G")))
            return("S")
        if (all(NT == c("C", "T")))
            return("Y")
        if (all(NT == c("G", "T")))
            return("K")
    } else if (length(NT) == 3) {
        if (all(NT == c("A", "C", "G")))
            return("V")
        if (all(NT == c("A", "C", "T")))
            return("H")
        if (all(NT == c("A", "G", "T")))
            return("D")
        if (all(NT == c("C", "G", "T")))
            return("B")
    } else {
        if (all(NT == c("A", "C", "G", "T")))
            return("N")
        
    }
    
    stop("wrong alphabet...")
}

##' convert fasta file to interleaved format
##'
##' 
##' @title fa_to_interleaved
##' @rdname fa_conversion
##' @param file fasta file
##' @param outfile output file
##' @return None
##' @export
##' @author Guangchuang Yu
##' @examples
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' fa1 <- tempfile(fileext = '.fa')
##' fa2 <- tempfile(fileext = '.fa')
##' fa_to_interleaved(fa_file, fa1)
##' fa_to_sequential(fa_file, fa2)
fa_to_interleaved <- function(file, outfile) {
    fa_write(readBStringSet(file), outfile, type="interleaved")
}

##' convert fasta file to sequential format
##'
##' 
##' @title fa_to_sequential
##' @rdname fa_conversion
##' @param file fasta file
##' @param outfile output file
##' @return None
##' @export
##' @author Guangchuang Yu
fa_to_sequential <- function(file, outfile) {
    fa_write(readBStringSet(file), outfile, type="sequential")
}

##' sequence alignment
##'
##' 
##' @title bs_aln
##' @param x XStringSet object
##' @param method alignment method
##' @param ... additional parameter
##' @return aligned sequences, XStringSet object
##' @importFrom Biostrings DNAStringSet
##' @importFrom Biostrings RNAStringSet
##' @importFrom Biostrings AAStringSet
##' @export
##' @author Guangchuang Yu
##' @examples
##' \dontrun{
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' bs_aln(x)
##' }
bs_aln <- function(x, method = 'muscle', ...) {
    method <- match.arg(method, c('muscle'))
    if (method == 'muscle') {
        res <- muscle::muscle(x, ...)
    }

    switch(class(res),
           DNAMultipleAlignment = DNAStringSet(res),
           RNAMultipleAlignment = RNAStringSet(res),
           AAMultipleAlignment = AAStringSet(res)
           )
}

getPhyInfo <- function(phy) {
    line1 <- readLines(phy, n=1)
    res <- strsplit(line1, split="\\s")[[1]]
    res <- res[res != ""]

    return(list(num=as.numeric(res[1]), width=as.numeric(res[2])))
}

##' convert bam file to aligned fasta file
##'
##'
##' @title bam2DNAStringSet
##' @param bamfile bam file
##' @param refseq refseq, DNAStringSet object
##' @return DNAStringSet object
##' @export
##' @author Guangchuang Yu
bam2DNAStringSet <- function(bamfile, refseq) {
    w <- width(refseq)
    if (length(refseq) > 1) {
        stop("please use bam2DNAStringSet2...")
    }
    pm <- GenomicRanges::GRanges(names(refseq), IRanges::IRanges(1, w))

    GenomicAlignments::stackStringsFromBam(Rsamtools::BamFile(bamfile),
                                          param=pm,
                                          Lpadding.letter='-',
                                          Rpadding.letter='-',
                                          use.names=TRUE)

}

##' convert bam file to aligned fasta file
##'
##'
##' @title bam2DNAStringSet2
##' @param bamfile bam file
##' @param refseq refseq, DNAStringSet object
##' @return list of DNAStringSet objects
##' @export
##' @author Guangchuang Yu
bam2DNAStringSet2 <- function(bamfile, refseq) {
    lapply(seq_along(refseq), function(i) {
        bam2DNAStringSet(bamfile, refseq[i])
    })
}

utils::globalVariables(".")

## open a file connection, transparently handling compressed files (.gz/.bz2/.xz)
.seq_connection <- function(filename, mode="r") {
    if (grepl("\\.gz$", filename, ignore.case=TRUE)) {
        gzfile(filename, mode)
    } else if (grepl("\\.bz2$", filename, ignore.case=TRUE)) {
        bzfile(filename, mode)
    } else if (grepl("\\.xz$", filename, ignore.case=TRUE)) {
        xzfile(filename, mode)
    } else {
        file(filename, mode)
    }
}

## return a plain text file path, decompressing .bz2/.xz to a temp file
## (Biostrings only reads gzip-compressed files natively)
.plain_text_file <- function(filename) {
    ext <- tolower(sub(".*\\.", "", filename))
    if (!ext %in% c("bz2", "xz")) {
        return(filename)
    }
    con <- .seq_connection(filename, "r")
    on.exit(close(con))
    txt <- tempfile()
    out <- file(txt, "w")
    on.exit(close(out), add=TRUE)
    writeLines(readLines(con), out)
    return(txt)
}

##' get id at specific position
##'
##'
##' @title get_id
##' @param x sequence description line
##' @param sep separator to split x
##' @param position id position
##' @return id
##' @export
##' @author Guangchuang Yu
##' @examples
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' get_id(names(x)[1:5], sep = " ", position=1)
get_id <- function(x, sep=" ", position) {
    res <- tokenize(x, sep)
    sapply(res, function(x) x[position])
}

##' sequence length
##'
##'
##' @title seqlen
##' @param fasfile fasta file
##' @return numeric vector
##' @export
##' @author Guangchuang Yu
seqlen <- function(fasfile) {
  x <- fa_read(fasfile)
  yy <- toString(x)
  yy <- gsub("-", "", yy)
  yy <- gsub(" ", "", yy)
  xx <- strsplit(yy, ",")
  sapply(xx, nchar)
}


##' summarize sequences in a fasta file
##'
##'
##' Compute number of sequences, length distribution, GC content and
##' proportion of ambiguous characters (anything other than A/C/G/T/U,
##' including gaps).
##' @title fa_summary
##' @param x fasta file or XStringSet object
##' @return data.frame with one row: \code{n_seq}, \code{min_len},
##' \code{max_len}, \code{mean_len}, \code{GC_pct} and \code{ambiguous_pct}
##' @importFrom Biostrings width
##' @export
##' @author Guangchuang Yu
##' @examples
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' fa_summary(fa_file)
fa_summary <- function(x) {
    if (is.character(x)) {
        x <- fa_read(x)
    }
    lens <- width(x)
    ch <- unlist(strsplit(toupper(as.character(x)), "", fixed=TRUE),
                 use.names=FALSE)
    clean <- ch %in% c("A", "C", "G", "T", "U")
    denom <- sum(clean)
    gc_pct <- if (denom == 0) NA_real_
              else round(100 * sum(ch %in% c("G", "C")) / denom, 2)

    data.frame(n_seq=length(x),
               min_len=min(lens),
               max_len=max(lens),
               mean_len=round(mean(lens), 2),
               GC_pct=gc_pct,
               ambiguous_pct=round(100 * mean(!clean), 2))
}

##' download genbank or fasta file by accession number
##'
##'
##' @title download_genbank
##' @param acc accession number(s)
##' @param db supported db, currently 'nuccore'
##' @param format one of 'genbank' or 'fasta'
##' @param outfile output file, by default, acc.gb or acc.fa
##' @return output file vector
##' @export
##' @author Guangchuang Yu
##' @examples
##' \dontrun{
##' tmpgb <- tempfile(fileext = '.gb')
##' tmpfa <- tempfile(fileext = '.fa')
##' download_genbank(acc='AB115403', format='genbank', outfile=tmpgb)
##' download_genbank(acc='AB115403', format='fasta', outfile=tmpfa)
##' }
download_genbank <- function(acc, db="nuccore", format = "genbank", outfile=NULL) {
    if (!is.null(outfile) && (length(outfile) != length(acc))) {
        stop("'outfile' & 'acc' should be equal length...")
    }
    acc <- gsub("\\.\\d+", "", acc)

    ofs <- character(length(acc))
    for (i in seq_along(acc)) {
        if (is.null(outfile)) {
            of <- NULL
        } else {
            of <- outfile[i]
        }
        ofs[i] <- download_genbank_item(acc[i], db, format, of)
    }
    invisible(ofs)
}

##' @importFrom utils download.file
download_genbank_item <- function(acc, db="nuccore", format = "genbank", outfile=NULL) {
    format <- match.arg(format, c('genbank', 'fasta'))
    if (db != "nuccore") {
        stop("currently, only nuccore is supported...")
    }
    url <- paste0("https://www.ncbi.nlm.nih.gov/sviewer/viewer.cgi?tool=portal&sendto=on&log$=seqview&db=",
                  db, '&dopt=', format, '&sort=&val=', acc)

    if (is.null(outfile)) {
        suffix <- '.gb'
        if (format == 'fasta')
            suffix <- '.fa'
        outfile <- paste0(acc, suffix)
    }
    
    yulab.utils:::mydownload(url=url, destfile = outfile)
    invisible(outfile)
}

##' download fasta sequence by accession number and read as sequence object
##'
##'
##' @title ncbi_fa_read
##' @param acc accession number(s)
##' @param db supported db, currently 'nuccore'
##' @param type one of 'DNA', 'RNA', 'AA', 'unknown' or 'auto'
##' @return BStringSet object
##' @export
##' @author Guangchuang Yu
##' @examples
##' \dontrun{
##' x <- ncbi_fa_read(c('AB115403', 'AB115403.1'))
##' x
##' }
ncbi_fa_read <- function(acc, db="nuccore", type = "auto") {
    if (length(acc) == 0) {
        stop("'acc' is empty...")
    }
    if (anyNA(acc)) {
        stop("'acc' contains NA...")
    }

    if (db != "nuccore") {
        stop("currently, only nuccore is supported...")
    }

    acc <- as.character(acc)
    acc <- gsub("\\.\\d+", "", acc)

    acc_query <- paste0(acc, collapse=",")
    url <- paste0(
        "https://www.ncbi.nlm.nih.gov/sviewer/viewer.cgi?tool=portal&sendto=on&log$=seqview&db=",
        db,
        "&dopt=fasta&sort=&val=",
        acc_query
    )

    f <- tempfile(fileext = ".fasta")
    yulab.utils:::mydownload(url=url, destfile=f)
    fa_read(f, type=type)
}

##' combine 2 fasta files into 1
##'
##' 
##' @title fa_combine
##' @param file1 fasta file 1
##' @param file2 fasta file 2
##' @param outfile output file
##' @param type one of interleaved and sequential
##' @return BStringSet
##' @export
##' @author Guangchuang Yu
##' @examples
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' fa1 <- tempfile(fileext=".fa")
##' fa2 <- tempfile(fileext=".fa")
##' fa_write(x[1:5], fa1)
##' fa_write(x[6:10], fa2)
##' fa_combine(fa1, fa2)
fa_combine <- function(file1, file2, outfile=NULL, type="interleaved") {
    fa1 <- readBStringSet(file1)
    fa2 <- readBStringSet(file2)
    fa <- c(fa1, fa2)

    if (!is.null(outfile)) {
        fa_write(fa, outfile, type)
    }
    invisible(fa)
}

##' convert fasta (aligned sequences) to phylip format
##'
##' 
##' @title fas2phy 
##' @param fasfile aligned sequences in fasta format
##' @param outfile output file
##' @param type one of interleaved and sequential
##' @return None
##' @export
##' @author Guangchuang Yu
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' phy_file <- tempfile(fileext = ".phy")
##' fas2phy(fa_file, phy_file)
fas2phy <- function(fasfile, outfile="out.phy", type="sequential") {
    x <- fa_read(fasfile)
    phy_write(x, outfile, type)
}

##' hamming distances of sequences
##'
##'
##' @title bs_hamming
##' @param x BStringSet object
##' @param count_indel whether count indel or not
##' @param ... additional parameters
##' @return hamming distance
##' @importFrom utils setTxtProgressBar
##' @importFrom utils txtProgressBar
##' @export
##' @author Guangchuang Yu
##' @examples
##' \dontrun{
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' ## align first 5 sequences, use `bs_aln(x)` to align all sequences
##' aln <- bs_aln(x[1:5])
##' bs_hamming(aln)
##' }
bs_hamming <- function(x, count_indel=FALSE, ...) {
    n <- length(x)
    if (n < 2) {
        stop("at least 2 sequences was expected...")
    }

    if ( n == 2 ) {
        s1 <- toString(x[1])
        s2 <- toString(x[2])
        res <- str_hamming(s1, s2, count_indel, ...)
        return(res)
    }

    res <- matrix(NA, nrow=n, ncol=n)
    colnames(res) <- rownames(res) <- names(x)
    seqs <- lapply(x, function(xx) {
        y <- toString(xx)
        substring(y, 1:nchar(y), 1:nchar(y))
    })
    cnt <- 1
    pb <- txtProgressBar(min=0, max=sum(1:n), style=3)

    for (i in 1:n) {
        for (j in 1:i) {
            setTxtProgressBar(pb, cnt)
            cnt <- cnt + 1
            if (i == j) {
                res[i,j] <- 0
            } else {
                res[i,j] <- suppressMessages(charVector_hamming(seqs[[i]], seqs[[j]], count_indel, ...))
                res[j, i] <- res[i, j]
            }
        }
    }
    close(pb)
    return(res)
}

str_hamming <- function(seq1, seq2, count_indel=FALSE, ...) {
    if (length(seq1) == 1 & length(seq2) == 1) {
        n1 <- nchar(seq1)
        n2 <- nchar(seq2)
        if (n1 != n2) {
            stop("length of two sequences should be consistent...")
        }
        seq1 <- substring(seq1, 1:n1, 1:n1)
        seq2 <- substring(seq2, 1:n2, 1:n2)
    }

    if ( length(seq1) > 1 & length(seq2) > 1) {
        charVector_hamming(seq1, seq2, count_indel, ...)
    } else {
        stop("seq1 and seq2 should be character of equal length...")
    }
}


charVector_hamming <- function(seq1, seq2, count_indel, type = "mapping", threshold=0.3) {
    type <- match.arg(type, c("mapping", "alignment"))
    ## mapping for mapping short read to long sequence
    ## alignment for aligning sequences of similar lengths

    ## for mapping
    ##
    ##
    ## gap will not be counted
    ##
    ## ---------XXXXXXX----------
    ## XXXXXXXXXXXXXXXXXXXXXXXXXX
    ##
    ## gap should be counted
    ##
    ## -------------------XXXXXXX
    ## XXXXXXXXXXXXXXXXXXXXX-----
    ##

    if (type == "mapping") {
        n1 <- seq1 != '-'
        n2 <- seq2 != '-'
        nn <- min(sum(n1), sum(n2))
        if (sum(n1 & n2)/nn < threshold) {
            count_indel <- TRUE
        }
    }

    ii <- which(seq1 != seq2)

    if (count_indel == FALSE) {
        message("--> indel will not count...")
        ii <- ii[seq1[ii] != '-' & seq2[ii] != '-']
    }

    return(length(ii))
}

##' rename fasta sequence name
##'
##'
##' @title fa_rename
##' @param fasfile fasta file
##' @param mapping_file mapping file
##' @param position rename token at specific position
##' @param sep sepator to divide token
##' @param mode one of 'replace', 'prefix' or 'suffix'
##' @param outfile output file
##' @return BStringSet object
##' @importFrom utils read.delim
##' @export
##' @author Guangchuang Yu
fa_rename <- function(fasfile, mapping_file, position=NULL, sep=NULL, mode="replace", outfile) {
    x <- fa_read(fasfile)
    mapping <- read.delim(mapping_file, header=FALSE, stringsAsFactors=FALSE)
    y <- bs_rename(x, mapping, position, sep, mode)
    if (!missing(outfile)) {
        fa_write(y, outfile)
    }
    invisible(y)
}

##' rename sequence
##'
##'
##' @title bs_rename
##' @param x BStringSet object
##' @param mapping two column data.frame
##' @param position rename token at specific position
##' @param sep sepator to divide token
##' @param mode one of 'replace', 'prefix' or 'suffix'
##' @return BStringSet
##' @export
##' @author Guangchuang Yu
bs_rename <- function(x, mapping, position=NULL, sep=NULL, mode="replace") {
    mode <- match.arg(mode, c("replace", "prefix", "suffix"))

    nn <- names(x)

    if (is.null(sep)) {
        id <- names(x)
    } else {
        id <- sapply(tokenize(nn, sep), function(y) y[1])
    }

    if (!is.null(position)) {
        if (is.null(sep))
            stop("separator should be provided")

        tk <- sapply(tokenize(nn, sep), function(y) y[position])

        i <- match(id, mapping[,1])

        if (mode == "replace") {
            to <- mapping[i,2]
        } else if (mode == "prefix") {
            to <- paste(mapping[i, 2], tk, sep="")
        } else {
            to <- paste(tk, mapping[i, 2], sep="")
        }

        tks <- tokenize(nn, sep)
        tks <- lapply(seq_along(tks), function(i) {
            tks[[i]][position] <- to[i]
            tks[[i]]
        })

        id2 <- sapply(tks, paste0, collapse=sep)

    } else {
        i <- match(id, mapping[,1])
        if (mode == "replace") {
            id2 <- mapping[i, 2]
        } else if (mode == "prefix") {
            id2 <- paste(mapping[i,2], id, sep="")
        } else {
            id2 <- paste(id, mapping[i, 2], sep="")
        }
    }
    id2 <- as.character(id2)
    ## unmatched entries may embed 'NA' (with position) or be NA outright
    ii <- is.na(id2) | is.na(i)
    if (any(ii)) {
        message(sum(ii), " sequences failed to rename, original name will be kept...")
        id2[ii] <- names(x)[ii]
    }
    names(x) <- id2
    return(x)
}

tokenize <- function(x, sep=" ") {
    lapply(x, function(y) unlist(strsplit(y, split=sep)))
}

library(seqmagick)

x <- Biostrings::BStringSet(c("s1"="AAAA", "s2"="CCCC", "s3"="GGGG"))
mapping <- data.frame(V1=c("s1", "s2"),
                      V2=c("human", "mouse"))

test_that("bs_rename replaces matching names and keeps unmatched ones", {
    z <- expect_message(bs_rename(x, mapping), "failed to rename")
    expect_identical(names(z), c("human", "mouse", "s3"))
    expect_identical(unname(as.character(z)), unname(as.character(x)))
})

test_that("bs_rename prefix/suffix modes", {
    p <- bs_rename(x[1:2], mapping, mode="prefix")
    expect_identical(names(p), c("humans1", "mouses2"))

    s <- bs_rename(x[1:2], mapping, mode="suffix")
    expect_identical(names(s), c("s1human", "s2mouse"))
})

test_that("bs_rename with sep and position", {
    y <- Biostrings::BStringSet(c("s1 v2"="AAAA", "s2 v2"="CCCC", "s3 v2"="GGGG"))
    m <- data.frame(V1=c("s1", "s2"), V2=c("human", "mouse"))

    ## replace token at position 2
    z <- bs_rename(y, m, sep=" ", position=2)
    expect_identical(names(z), c("s1 human", "s2 mouse", "s3 v2"))

    ## replace token at position 1
    z1 <- bs_rename(y, m, sep=" ", position=1)
    expect_identical(names(z1), c("human v2", "mouse v2", "s3 v2"))
})

test_that("position requires separator", {
    expect_error(bs_rename(x, mapping, position=1),
                 "separator should be provided")
})

test_that("fa_rename end-to-end writes output file", {
    fasfile <- tempfile(fileext=".fas")
    fa_write(x, fasfile)

    mapfile <- tempfile()
    write.table(mapping, mapfile, sep="\t", quote=FALSE,
                row.names=FALSE, col.names=FALSE)

    outfile <- tempfile(fileext=".fas")
    z <- invisible(fa_rename(fasfile, mapfile, outfile=outfile))
    expect_identical(names(z)[1:2], c("human", "mouse"))

    reread <- fa_read(outfile)
    expect_identical(names(reread), c("human", "mouse", "s3"))

    unlink(c(fasfile, mapfile, outfile))
})

##' rename txt file (eg Description line of fasta file)
##' according to first token (eg accession number)
##'
##'
##' @title renameTXT
##' @param txt_file txt file
##' @param name_file name file
##' @param sep separator
##' @param split logical, split result or not
##' @return None
##' @export
##' @author Guangchuang Yu
renameTXT <- function(txt_file, name_file, sep="_", split=TRUE) {
    txt <- readLines(txt_file)
    res <- renameTXT_internal(txt, name_file, sep)
    if (split) {
        f1 <- file(gsub("\\.", "_IN\\.", txt_file), "w")
        writeLines(res$name[res$ii], f1)
        close(f1)
        f2 <- file(gsub("\\.", "_OUT\\.", txt_file), "w")
        writeLines(res$name[res$jj], f2)
        close(f2)
    } else {
        outfile <- file(gsub("\\.", "_rename\\.", txt_file), "w")
        writeLines(res$name, outfile)
        close(outfile)
    }
}

renameTXT_internal <- function(txt, name_file, sep="_") {
    sep <- paste0("\\", sep)
    acc_txt <- get_first_token(gsub("^>", "", txt), sep)

    nm <- readLines(name_file)
    nm <- unique(nm)
    acc <- get_first_token(gsub("^>", "", nm), sep)

    ii <- jj <- c()
    for (i in seq_along(acc_txt)) {
        j <- which(acc_txt[i] == acc)
        if (length(j) == 1) {
            ii <- c(ii, i)
            txt[i] <- nm[j]
        } else if (length(j) > 1) {
            ii <- c(ii, i)
            txt[i] <- nm[j[1]]
            warning("multiple match, first record will be used...\n", paste0("\t", nm[j], "\n"))
        } else {
            jj <- c(jj, i)
        }
    }
    res <- list(name=txt, ii=ii, jj=jj)
    return(res)
}

get_first_token <- function(x, sep) {
    trimws(sapply(x, function(y) unlist(strsplit(y, sep))[1]))
}

##' fasta filter by searching pattern
##'
##' 
##' @title fa_filter
##' @param fasfile input fasta file
##' @param pattern keyword for filter
##' @param by one of 'description' and 'sequence'
##' @param ignore.case logical
##' @param outfile output file
##' @param type one of 'interleaved' and 'sequential'
##' @return BStringSet object
##' @importFrom Biostrings readBStringSet
##' @export
##' @author Guangchuang Yu
fa_filter <- function(fasfile, pattern, by='description', ignore.case=FALSE,
                      outfile=NULL, type="interleaved") {
    
    fas <- readBStringSet(fasfile)

    res <- bs_filter(fas, pattern, by, ignore.case)
    
    if (!is.null(outfile)) {
       fa_write(res, outfile, type)
    }
    
    invisible(res)
}

##' biological sequence filter by searching pattern
##'
##' 
##' @title bs_filter
##' @param x BStringSet object
##' @param pattern keyword for filter
##' @param by one of 'description' and 'sequence'
##' @param ignore.case logical
##' @return BStringSet object
##' @importFrom Biostrings matchPattern
##' @export
##' @author Guangchuang Yu
##' @examples
##' fa_file <- system.file("extdata/HA.fas", package="seqmagick")
##' x <- fa_read(fa_file)
##' bs_filter(x, 'ATGAAAGTAAAA', by='sequence')
bs_filter <- function(x, pattern, by='description', ignore.case=FALSE) {
    by <- match.arg(by, c("description", "sequence"))
    
    if (by == 'description') {
        desc <- names(x)
        if (ignore.case) {
            pattern <- toupper(pattern)
            desc <- toupper(desc)
        }
        idx <- grep(pattern, desc)
    } else {
        if (ignore.case) {
            hit <- sapply(x, function(y) matchPattern(toupper(pattern), toupper(y)))
        } else {
            hit <- sapply(x, function(y) matchPattern(pattern, y))
        }
        idx <- which(sapply(hit, length) > 0)
    }

    if (length(idx) == 0) {
        message("pattern not found...")
        return(NULL)
    }
    
    return(x[idx])
}

##' convert phylip file to fasta file
##'
##' 
##' @title phy2fas
##' @param phyfile phylip file 
##' @param outfile output file
##' @param type one of interleaved and sequential
##' @return None
##' @export
##' @author Guangchuang Yu
##' @examples
##' phy_file <- system.file("extdata/HA.phy", package="seqmagick")
##' fa_file <- tempfile(fileext = '.fas')
##' phy2fas(phy_file, fa_file)
phy2fas <- function(phyfile, outfile="out.fas", type="interleaved") {
    x <- phy_read(phyfile)
    fa_write(x, outfile, type)
}

##' replace character for example from '-' to 'N' of fasta sequence that only applied inside sequence
##' any '-' character at start/end of the sequence (aligned seqs may contains '-' at prefix/suffix) will not be replaced
##' 
##'
##' 
##' @title replaceInside
##' @param fasfile fasta file
##' @param from character to be replaced, '-' by default
##' @param to replace 'from' to 'to', 'N' by default
##' @param outfile output file
##' @importFrom Biostrings readDNAStringSet
##' @importFrom Biostrings writeXStringSet
##' @export
##' @return DNAStringSet
##' @author Guangchuang Yu
replaceInside <- function(fasfile, from="-", to="N", outfile=NULL) {
    fas <- readDNAStringSet(fasfile)
    fas_replaced <- replaceInside.DNAStringSet(fas, from, to)
    if (is.null(outfile)) {
        outfile <- sub("\\.(fa.*)$", "_replaced.\\1", fasfile)
    }
    writeXStringSet(fas_replaced, outfile)
    invisible(fas_replaced)
}

##' @importFrom Biostrings DNAStringSet
replaceInside.DNAStringSet <- function(x, from, to) {
    strings <- sapply(seq_along(x), function(i)
        replaceInside.character(toString(x[i]), from=from, to=to))
    names(strings) <- names(x)
    DNAStringSet(strings)
}


replaceInside.character <- function(x, from="-", to="N") {
    ## pattern matches start/end will be ignore
    ##
    ## > replaceInside.character("----ABC--ABDC---EDFFF---")
    ## [1] "----ABCNNABDCNNNEDFFF---"
    
    y <- gsub(from, to, x)
    
    x <- gsub(paste0("^", to), from, y)

    ## restore start
    while (x != y) {
        y <- x
        x <- gsub(paste0("^(", from, "*)", to), paste0("\\1", from), y)
    }

    ## restore end
    y <- x
    x <- gsub(paste0(to, "$"), from, y)
    while(x != y) {
        y <- x
        x <- gsub(paste0(to, "(", from, "*)$"), paste0(from, "\\1"), y)
    }

    return(x)
}

# Personal spelling dictionary for `R CMD check --as-cran`.
# Keeps the default 'en_stats' dictionary and adds package-specific
# technical terms so that the DESCRIPTION spell check does not flag
# them as "possibly misspelled words".
#
# The word list used to build seqmagick.rds is:
#   GC, gzip, bzip, xz, FASTA, PHYLIP, CLUSTAL, STOCKHOLM, MEGA,
#   GenBank, NCBI, Biostrings, seqmagick
description <- list(
    dictionaries = c("en_stats", "seqmagick")
)

#!/bin/bash


seqmagick convert --apply-function myfunctions.py:no_gaps \
  ../aligned.fasta empty.fasta

seqmagick convert --apply-function myfunctions.py:hash_starts_numeric \
  ../aligned.fasta hashed.fasta

---
title: "seqmagick: sequence manipulation"
author: 
- name: Guangchuang Yu
  email: guangchuangyu@gmail.com
  affiliation: Department of Bioinformatics, School of Basic Medical Sciences, Southern Medical University
date: "`r Sys.Date()`"
bibliography: seqmagick.bib
biblio-style: apalike
output:
  prettydoc::html_pretty:
    toc: true
    highlight: github
    theme: cayman
  pdf_document:
    toc: true
vignette: >
  %\VignetteIndexEntry{seqmagick introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\usepackage[utf8]{inputenc}
---

```{r style, echo=FALSE, results="asis", message=FALSE}
knitr::opts_chunk$set(tidy = FALSE,
		   message = FALSE)
```


```{r echo=FALSE, results="hide", message=FALSE}
library(magrittr)
library(Biostrings)
library("seqmagick")
```

# Download sequences

## Genbank

```{r eval=FALSE}
tmpgb <- tempfile(fileext = '.gb')
tmpfa <- tempfile(fileext = '.fa')
download_genbank(acc='AB115403', format='genbank', outfile=tmpgb)
download_genbank(acc='AB115403', format='fasta', outfile=tmpfa)
## readLines(tmpgb)[1:10]
## readLines(tmpfa)
```

# File conversion

## fasta and phylip conversion

```{r eval=F}
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
## use the small subset to save compilation time of the vignette
fa2 <- tempfile(fileext = '.fa')
fa_read(fa_file) %>% bs_filter('ATGAAAGTAAAA', by='sequence') %>% fa_write(fa2, type='interleaved')


alnfas <- tempfile(fileext = ".fas")
fa_read(fa2) %>% bs_aln(quiet=TRUE) %>% fa_write(alnfas)

## phylip format is only for aligned sequences
tmpphy <- tempfile(fileext = ".phy")
fas2phy(alnfas, tmpphy, type = 'sequential')
```

`seqmagick` supports both `sequential` and `interleaved` formats, users can specify the format by `type` parameter.


```{r eval=F}
phy2fas(tmpphy, alnfas, type = 'interleaved')
```


## interleaved and sequential format conversion



```{r eval=F}
tmpfas <- tempfile(fileext='.fa')
fa_read(fa2) %>% fa_write(tmpfas, type="sequential")
```

```{r eval=F} 
tmpphy2 <- tempfile(fileext = '.phy')
phy_read(tmpphy) %>% phy_write(tmpphy2, type="interleaved")
```

# Sequence manipulation


```{r eval=F}
bs <- fa_read(fa_file)
bs_filter(bs, 'ATGAAAGTAAAA', by='sequence')

aln <- bs_filter(bs, 'ATGAAAGTAAAA', by='sequence') %>% bs_aln(quiet=TRUE)

bs_consensus(aln)
```



# Bugs/Feature requests

If you have any, [let me know](https://github.com/YuLab-SMU/seqmagick/issues). Thx!

# Session info

Here is the output of `sessionInfo()` on the system on which this document was compiled:
```{r echo=FALSE}
sessionInfo()
```


# References

---
title: "seqmagick: sequence manipulation"
author: 
- name: Guangchuang Yu
  email: guangchuangyu@gmail.com
  affiliation: Department of Bioinformatics, School of Basic Medical Sciences, Southern Medical University
date: "`r Sys.Date()`"
bibliography: seqmagick.bib
biblio-style: apalike
output:
  prettydoc::html_pretty:
    toc: true
    highlight: github
    theme: cayman
  pdf_document:
    toc: true
vignette: >
  %\VignetteIndexEntry{seqmagick introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\usepackage[utf8]{inputenc}
---

```{r style, echo=FALSE, results="asis", message=FALSE}
knitr::opts_chunk$set(tidy = FALSE,
		   message = FALSE)
```


```{r echo=FALSE, results="hide", message=FALSE}
library(yulab.utils)
library(Biostrings)
library("seqmagick")
```

# Supported file formats

`seqmagick` works with `r Biocpkg("Biostrings")` objects and provides readers/writers for a set of widely used sequence file formats. Compressed files (`gzip`, `bzip2` and `xz`) are supported transparently &mdash; just use them like plain text files.

| Format | Read | Write | Related functions |
|:-----------|:----------|:----------|:----------------------------|
| FASTA (.fas/.fa/.fasta) | `fa_read()` | `fa_write()` | `fa_to_interleaved()`, `fa_to_sequential()`, `fa_combine()`, `fas2phy()`, `phy2fas()` |
| PHYLIP (.phy) | `phy_read()` | `phy_write()` | `fas2phy()`, `phy2fas()` |
| CLUSTAL (.clw) | `clw_read()` | - | - |
| STOCKHOLM (.sth) | `sth_read()` | - | - |
| MEGA (.meg, .mega) | `mega_read()` | - | - |
| GenBank (.gb) | `gb_read()` | - | `download_genbank()` |
| NCBI (online) | `ncbi_fa_read()` | - | `download_genbank()` |
| BAM (.bam) | `bam2DNAStringSet()` | - | - |

FASTA and PHYLIP support both `sequential` and `interleaved` layouts; specify via the `type` parameter where applicable.

# Sequence I/O

## FASTA

```{r}
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
x
```

Compressed files can be written and read directly:

```{r}
tmpgz <- tempfile(fileext=".fas.gz")
fa_write(x[1:5], tmpgz)          ## gzip is used automatically
y <- fa_read(tmpgz)
identical(width(y), width(x[1:5]))
```

## PHYLIP

```{r}
phy_file <- system.file("extdata/HA.phy", package="seqmagick")
p <- phy_read(phy_file)
```

## CLUSTAL and STOCKHOLM

```{r eval=FALSE}
x <- clw_read("alignment.clw")
x <- sth_read("alignment.sth")
```

## MEGA

```{r}
meg <- system.file("extdata/mega/Crab_rRNA.meg", package="seqmagick")
m <- mega_read(meg)
m
```

## GenBank

```{r eval=FALSE}
gb <- system.file("extdata/AB115403.gb", package="seqmagick")
x <- gb_read(gb)
```

# Download sequences from NCBI

Both `download_genbank()` (writes to files) and `ncbi_fa_read()` (returns `r Biocpkg("Biostrings")` objects directly) fetch records by accession number.

```{r eval=FALSE}
## save to local files
download_genbank(acc='AB115403', format='genbank', outfile='out.gb')
download_genbank(acc='AB115403', format='fasta', outfile='out.fa')

## or read into R directly
x <- ncbi_fa_read(acc=c('AB115403', 'CY084969'))
```

# Format conversion

## fasta and phylip conversion

Note that PHYLIP is only meaningful for aligned sequences. We first subset the FASTA file, align it with `bs_aln()`, then convert it to PHYLIP:

```{r eval=FALSE}
## use a small subset to keep this demo fast
fa2 <- tempfile(fileext = '.fa')
fa_read(fa_file) |> bs_filter('ATGAAAGTAAAA', by='sequence') |> fa_write(fa2, type='interleaved')

alnfas <- tempfile(fileext = ".fas")
fa_read(fa2) |> bs_aln(quiet=TRUE) |> fa_write(alnfas)

tmpphy <- tempfile(fileext = ".phy")
fas2phy(alnfas, tmpphy, type = 'sequential')
```

Converting back from PHYLIP to FASTA is symmetric:

```{r eval=FALSE}
phy2fas(tmpphy, alnfas, type = 'interleaved')
```

## interleaved and sequential format conversion

Use the `type` parameter in `fa_write()`/`phy_write()`, or the shortcuts `fa_to_interleaved()`/`fa_to_sequential()`:

```{r eval=FALSE}
## inter-convert by read + write with different 'type'
fa_read(fa2) |> fa_write("out.fas", type="sequential")
phy_read(tmpphy) |> phy_write("out.phy", type="interleaved")

## or one-liner conversions for FASTA files
fa_to_interleaved("in.fas", "interleaved.fas")
fa_to_sequential("in.fas", "sequential.fas")
```

# Sequence manipulation

## Filtering, alignment and consensus

```{r eval=FALSE}
bs <- fa_read(fa_file)

## keep only sequences containing the pattern
f <- bs_filter(bs, 'ATGAAAGTAAAA', by='sequence')

## multiple sequence alignment requires the 'muscle' package
aln <- f |> bs_aln(quiet=TRUE)

## consensus sequence of the alignment
bs_consensus(aln)
```

## Renaming sequences

Rename sequences by supplying a two-column mapping (old name -> new name). Unmatched names are kept unchanged.

```{r}
mapping <- data.frame(old = c(names(x)[1], names(x)[2]),
                      new = c("HA_HK", "HA_BR"))
z <- bs_rename(x, mapping)
names(z)[1:4]
```

With `sep` and `position`, only a specific token of each name (defined by splitting the name with `sep`) is replaced by the mapped value: 

```{r}
m2 <- data.frame(old = c("seq1", "seq2"),
                 new = c("HA_human", "HA_swine"))
demo <- Biostrings::BStringSet(c("seq1 HK01"="AAAA",
                                 "seq2 TW02"="CCCC",
                                 "seq3 JP03"="GGGG"))
z2 <- bs_rename(demo, m2, sep=" ", position=1)
names(z2)
```

For files, `fa_rename()` reads a FASTA file plus a two-column table and writes renamed output in one step:

```{r eval=FALSE}
fa_rename("seqs.fas", "map.txt", outfile="renamed.fas")
```

# Summary statistics

The new `fa_summary()` gives a quick overview of a FASTA file (or an existing `r Biocpkg("Biostrings")` object): number of sequences, length range, GC content and proportion of ambiguous characters.

```{r}
fa_summary(fa_file)
```

Works on in-memory objects as well:

```{r}
fa_summary(x[1:10])
```

For per-sequence lengths (gap characters excluded), use the classic `seqlen()`:

```{r}
head(seqlen(fa_file))
```

# Bugs/Feature requests

If you have any, [let me know](https://github.com/YuLab-SMU/seqmagick/issues). Thx!

# Session info

Here is the output of `sessionInfo()` on the system on which this document was compiled:
```{r echo=FALSE}
sessionInfo()
```


# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fa_combine.R
\name{fa_combine}
\alias{fa_combine}
\title{fa_combine}
\usage{
fa_combine(file1, file2, outfile = NULL, type = "interleaved")
}
\arguments{
\item{file1}{fasta file 1}

\item{file2}{fasta file 2}

\item{outfile}{output file}

\item{type}{one of interleaved and sequential}
}
\value{
BStringSet
}
\description{
combine 2 fasta files into 1
}
\examples{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
fa1 <- tempfile(fileext=".fa")
fa2 <- tempfile(fileext=".fa")
fa_write(x[1:5], fa1)
fa_write(x[6:10], fa2)
fa_combine(fa1, fa2)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqmagick-package.R
\docType{package}
\name{seqmagick-package}
\alias{seqmagick}
\alias{seqmagick-package}
\title{seqmagick: Sequence Manipulation Utilities}
\description{
Tools for reading and writing biological sequences in multiple formats, including 'FASTA', 'PHYLIP', 'CLUSTAL', 'STOCKHOLM', 'MEGA' and 'GenBank'. Supports interleaved and sequential layouts where applicable, and reads or writes gzip, bzip2 and xz compressed files transparently. Converts between formats, renames sequences via a two-column mapping, summarizes sequence counts, lengths, GC content and ambiguous characters, and manipulates sequence sets (e.g., filtering by patterns and computing consensus sequences from alignments). Also includes functions to download nucleotide records from NCBI by accession.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/YuLab-SMU/seqmagick}
  \item Report bugs at \url{https://github.com/YuLab-SMU/seqmagick/issues}
}

}
\author{
\strong{Maintainer}: Guangchuang Yu \email{guangchuangyu@gmail.com} (\href{https://orcid.org/0000-0002-6485-8781}{ORCID})

Authors:
\itemize{
  \item Guangchuang Yu \email{guangchuangyu@gmail.com} (\href{https://orcid.org/0000-0002-6485-8781}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{phy_write}
\alias{phy_write}
\title{phy_write}
\usage{
phy_write(x, outfile, type = "sequential")
}
\arguments{
\item{x}{XStringSet object}

\item{outfile}{output file}

\item{type}{one of interleaved and sequential}
}
\value{
None
}
\description{
write phylip file
}
\examples{
\dontrun{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
aln <- bs_aln(x[1:5])
phy_file <- tempfile(fileext = '.phy')
phy_write(aln, phy_file)
}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{phy_read}
\alias{phy_read}
\title{phy_read}
\usage{
phy_read(file)
}
\arguments{
\item{file}{phylip file}
}
\value{
BStringSet object
}
\description{
read aligned sequences in phylip format
}
\examples{
phy_file <- system.file("extdata/HA.phy", package="seqmagick")
phy_read(phy_file)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renameTXT.R
\name{renameTXT}
\alias{renameTXT}
\title{renameTXT}
\usage{
renameTXT(txt_file, name_file, sep = "_", split = TRUE)
}
\arguments{
\item{txt_file}{txt file}

\item{name_file}{name file}

\item{sep}{separator}

\item{split}{logical, split result or not}
}
\value{
None
}
\description{
rename txt file (eg Description line of fasta file)
according to first token (eg accession number)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NCBI.R
\name{ncbi_fa_read}
\alias{ncbi_fa_read}
\title{ncbi_fa_read}
\usage{
ncbi_fa_read(acc, db = "nuccore", type = "auto")
}
\arguments{
\item{acc}{accession number(s)}

\item{db}{supported db, currently 'nuccore'}

\item{type}{one of 'DNA', 'RNA', 'AA', 'unknown' or 'auto'}
}
\value{
BStringSet object
}
\description{
download fasta sequence by accession number and read as sequence object
}
\examples{
\dontrun{
x <- ncbi_fa_read(c('AB115403', 'AB115403.1'))
x
}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NCBI.R
\name{download_genbank}
\alias{download_genbank}
\title{download_genbank}
\usage{
download_genbank(acc, db = "nuccore", format = "genbank", outfile = NULL)
}
\arguments{
\item{acc}{accession number(s)}

\item{db}{supported db, currently 'nuccore'}

\item{format}{one of 'genbank' or 'fasta'}

\item{outfile}{output file, by default, acc.gb or acc.fa}
}
\value{
output file vector
}
\description{
download genbank or fasta file by accession number
}
\examples{
\dontrun{
tmpgb <- tempfile(fileext = '.gb')
tmpfa <- tempfile(fileext = '.fa')
download_genbank(acc='AB115403', format='genbank', outfile=tmpgb)
download_genbank(acc='AB115403', format='fasta', outfile=tmpfa)
}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mega.R, R/read.R
\name{mega_read}
\alias{mega_read}
\alias{fa_read}
\alias{clw_read}
\alias{sth_read}
\title{read sequence alignment file}
\usage{
mega_read(file)

fa_read(file, type = "auto")

clw_read(file, type = "auto")

sth_read(file, type = "auto")
}
\arguments{
\item{file}{multiple sequence file, gzip/bzip2/xz compressed files are supported (.gz/.bz2/.xz)}

\item{type}{one of 'DNA', 'RNA', 'AA', 'Protein', 'unknown' or 'auto'}
}
\value{
BStringSet object
}
\description{
read sequence alignment file
}
\examples{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
fa_read(fa_file)

mega_file <- system.file("extdata/mega/Crab_rRNA.meg", package="seqmagick")
mega_read(mega_file)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aln.R
\name{bs_aln}
\alias{bs_aln}
\title{bs_aln}
\usage{
bs_aln(x, method = "muscle", ...)
}
\arguments{
\item{x}{XStringSet object}

\item{method}{alignment method}

\item{...}{additional parameter}
}
\value{
aligned sequences, XStringSet object
}
\description{
sequence alignment
}
\examples{
\dontrun{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
bs_aln(x)
}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fa_conversion.R
\name{fa_to_interleaved}
\alias{fa_to_interleaved}
\alias{fa_to_sequential}
\title{fa_to_interleaved}
\usage{
fa_to_interleaved(file, outfile)

fa_to_sequential(file, outfile)
}
\arguments{
\item{file}{fasta file}

\item{outfile}{output file}
}
\value{
None

None
}
\description{
convert fasta file to interleaved format

convert fasta file to sequential format
}
\examples{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
fa1 <- tempfile(fileext = '.fa')
fa2 <- tempfile(fileext = '.fa')
fa_to_interleaved(fa_file, fa1)
fa_to_sequential(fa_file, fa2)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bam2DNAStringSet.R
\name{bam2DNAStringSet2}
\alias{bam2DNAStringSet2}
\title{bam2DNAStringSet2}
\usage{
bam2DNAStringSet2(bamfile, refseq)
}
\arguments{
\item{bamfile}{bam file}

\item{refseq}{refseq, DNAStringSet object}
}
\value{
list of DNAStringSet objects
}
\description{
convert bam file to aligned fasta file
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqlen.R
\name{seqlen}
\alias{seqlen}
\title{seqlen}
\usage{
seqlen(fasfile)
}
\arguments{
\item{fasfile}{fasta file}
}
\value{
numeric vector
}
\description{
sequence length
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phy2fas.R
\name{phy2fas}
\alias{phy2fas}
\title{phy2fas}
\usage{
phy2fas(phyfile, outfile = "out.fas", type = "interleaved")
}
\arguments{
\item{phyfile}{phylip file}

\item{outfile}{output file}

\item{type}{one of interleaved and sequential}
}
\value{
None
}
\description{
convert phylip file to fasta file
}
\examples{
phy_file <- system.file("extdata/HA.phy", package="seqmagick")
fa_file <- tempfile(fileext = '.fas')
phy2fas(phy_file, fa_file)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{fa_write}
\alias{fa_write}
\title{fa_write}
\usage{
fa_write(x, outfile, type = "interleaved")
}
\arguments{
\item{x}{XStringSet object}

\item{outfile}{output file}

\item{type}{one of interleaved and sequential}
}
\value{
None
}
\description{
write fasta file
}
\examples{
phy_file <- system.file("extdata/HA.phy", package="seqmagick")
x <- phy_read(phy_file)
fa_file <- tempfile(fileext = '.fas')
fa_write(x, fa_file)
}
\references{
\url{https://phylipweb.github.io/phylip/}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hamming.R
\name{bs_hamming}
\alias{bs_hamming}
\title{bs_hamming}
\usage{
bs_hamming(x, count_indel = FALSE, ...)
}
\arguments{
\item{x}{BStringSet object}

\item{count_indel}{whether count indel or not}

\item{...}{additional parameters}
}
\value{
hamming distance
}
\description{
hamming distances of sequences
}
\examples{
\dontrun{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
## align first 5 sequences, use `bs_aln(x)` to align all sequences
aln <- bs_aln(x[1:5])
bs_hamming(aln)
}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename.R
\name{bs_rename}
\alias{bs_rename}
\title{bs_rename}
\usage{
bs_rename(x, mapping, position = NULL, sep = NULL, mode = "replace")
}
\arguments{
\item{x}{BStringSet object}

\item{mapping}{two column data.frame}

\item{position}{rename token at specific position}

\item{sep}{sepator to divide token}

\item{mode}{one of 'replace', 'prefix' or 'suffix'}
}
\value{
BStringSet
}
\description{
rename sequence
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter.R
\name{bs_filter}
\alias{bs_filter}
\title{bs_filter}
\usage{
bs_filter(x, pattern, by = "description", ignore.case = FALSE)
}
\arguments{
\item{x}{BStringSet object}

\item{pattern}{keyword for filter}

\item{by}{one of 'description' and 'sequence'}

\item{ignore.case}{logical}
}
\value{
BStringSet object
}
\description{
biological sequence filter by searching pattern
}
\examples{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
bs_filter(x, 'ATGAAAGTAAAA', by='sequence')
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replaceInside.R
\name{replaceInside}
\alias{replaceInside}
\title{replaceInside}
\usage{
replaceInside(fasfile, from = "-", to = "N", outfile = NULL)
}
\arguments{
\item{fasfile}{fasta file}

\item{from}{character to be replaced, '-' by default}

\item{to}{replace 'from' to 'to', 'N' by default}

\item{outfile}{output file}
}
\value{
DNAStringSet
}
\description{
replace character for example from '-' to 'N' of fasta sequence that only applied inside sequence
any '-' character at start/end of the sequence (aligned seqs may contains '-' at prefix/suffix) will not be replaced
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqlen.R
\name{fa_summary}
\alias{fa_summary}
\title{fa_summary}
\usage{
fa_summary(x)
}
\arguments{
\item{x}{fasta file or XStringSet object}
}
\value{
data.frame with one row: \code{n_seq}, \code{min_len},
\code{max_len}, \code{mean_len}, \code{GC_pct} and \code{ambiguous_pct}
}
\description{
summarize sequences in a fasta file
}
\details{
Compute number of sequences, length distribution, GC content and
proportion of ambiguous characters (anything other than A/C/G/T/U,
including gaps).
}
\examples{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
fa_summary(fa_file)
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{gb_read}
\alias{gb_read}
\title{gb_read}
\usage{
gb_read(file)
}
\arguments{
\item{file}{input genbank file}
}
\value{
sequence object
}
\description{
extract accession number and sequence from genbank file
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus.R
\name{consensus}
\alias{consensus}
\alias{bs_consensus}
\title{consensus}
\usage{
consensus(x, type = "DNA")

bs_consensus(x, type = "DNA", r = 1)
}
\arguments{
\item{x}{BStringSet object}

\item{type}{currently, only DNA supported}

\item{r}{if any NT > r, it will be selected as representative base}
}
\value{
consensus sequence string

consensus sequence string
}
\description{
consensus of aligned sequences

consensus of aligned sequences
}
\examples{
\dontrun{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
## align first 5 sequences, use `bs_aln(x)` to align all sequences
aln <- bs_aln(x[1:5])
## or bs_consensus(aln)
consensus(aln)
}
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter.R
\name{fa_filter}
\alias{fa_filter}
\title{fa_filter}
\usage{
fa_filter(
  fasfile,
  pattern,
  by = "description",
  ignore.case = FALSE,
  outfile = NULL,
  type = "interleaved"
)
}
\arguments{
\item{fasfile}{input fasta file}

\item{pattern}{keyword for filter}

\item{by}{one of 'description' and 'sequence'}

\item{ignore.case}{logical}

\item{outfile}{output file}

\item{type}{one of 'interleaved' and 'sequential'}
}
\value{
BStringSet object
}
\description{
fasta filter by searching pattern
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fas2phy.R
\name{fas2phy}
\alias{fas2phy}
\title{fas2phy}
\usage{
fas2phy(fasfile, outfile = "out.phy", type = "sequential")
}
\arguments{
\item{fasfile}{aligned sequences in fasta format}

\item{outfile}{output file}

\item{type}{one of interleaved and sequential}
}
\value{
None
}
\description{
convert fasta (aligned sequences) to phylip format
}
\author{
Guangchuang Yu
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
phy_file <- tempfile(fileext = ".phy")
fas2phy(fa_file, phy_file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename.R
\name{fa_rename}
\alias{fa_rename}
\title{fa_rename}
\usage{
fa_rename(
  fasfile,
  mapping_file,
  position = NULL,
  sep = NULL,
  mode = "replace",
  outfile
)
}
\arguments{
\item{fasfile}{fasta file}

\item{mapping_file}{mapping file}

\item{position}{rename token at specific position}

\item{sep}{sepator to divide token}

\item{mode}{one of 'replace', 'prefix' or 'suffix'}

\item{outfile}{output file}
}
\value{
BStringSet object
}
\description{
rename fasta sequence name
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_id.R
\name{get_id}
\alias{get_id}
\title{get_id}
\usage{
get_id(x, sep = " ", position)
}
\arguments{
\item{x}{sequence description line}

\item{sep}{separator to split x}

\item{position}{id position}
}
\value{
id
}
\description{
get id at specific position
}
\examples{
fa_file <- system.file("extdata/HA.fas", package="seqmagick")
x <- fa_read(fa_file)
get_id(names(x)[1:5], sep = " ", position=1)
}
\author{
Guangchuang Yu
}
